@@ -31,20 +31,22 @@
 import argparse
 import logging
 import os
+import subprocess
 import sys
 import tempfile
 
+import importlib_resources
 import pexpect
-import pkg_resources
 
-from PyQt4 import QtGui
-from PyQt4.QtCore import Qt
-from PyQt4.QtGui import QAction, QApplication, QDialog, QMainWindow, \
-    QMessageBox, QListWidgetItem
+from pexpect import fdpexpect
+
+from PySide2.QtCore import Qt
+from PySide2.QtWidgets import QAction, QApplication, \
+    QDialog, QDialogButtonBox, QMainWindow, QMessageBox, QListWidgetItem
 
 import egauge.ctid as ctid
 
-from egauge.qt4 import terminal
+from egauge.pyside2 import terminal
 
 from ctid_programmer import preferences
 from ctid_programmer import sensor_params
@@ -53,23 +55,18 @@ from ctid_programmer import sn_egauge
 from ctid_programmer import sn_local
 from ctid_programmer import template
 
-from ctid_programmer.qt4.main_window import Ui_MainWindow
-from ctid_programmer.qt4.template_dialog import Ui_Template_Dialog
-from ctid_programmer.qt4.preferences_dialog import Ui_Preferences_Dialog
-from ctid_programmer.qt4.param_ct import Ui_Param_CT
-from ctid_programmer.qt4.param_volt import Ui_Param_Volt
-from ctid_programmer.qt4.param_temp import Ui_Param_Temp
-from ctid_programmer.qt4.param_ntc import Ui_Param_NTC
-from ctid_programmer.qt4.param_pulse import Ui_Param_Pulse
+from ctid_programmer.gui.main_window import Ui_MainWindow
+from ctid_programmer.gui.template_dialog import Ui_Template_Dialog
+from ctid_programmer.gui.preferences_dialog import Ui_Preferences_Dialog
+from ctid_programmer.gui.param_ct import Ui_Param_CT
+from ctid_programmer.gui.param_volt import Ui_Param_Volt
+from ctid_programmer.gui.param_temp import Ui_Param_Temp
+from ctid_programmer.gui.param_ntc import Ui_Param_NTC
+from ctid_programmer.gui.param_pulse import Ui_Param_Pulse
 
 PATH_AVRDUDE = 'avrdude'
 PATH_CTID_ENCODER = 'ctid-encoder'
 
-PATH_RESOURCES = pkg_resources \
-                 .resource_filename(__name__,
-                                    'lib/CTid-programmer/resources')
-PATH_CODE_DIR = pkg_resources.resource_filename(__name__, 'resources/code')
-PATH_ICON_DIR = os.path.join(PATH_RESOURCES, 'icons')
 PATH_STATE_DIR = os.path.join(os.getenv('HOME'), '.CTid')
 
 SENSOR_LONG_NAME = {
@@ -94,16 +91,31 @@ CHIP_ID_TO_NAME = {
 }
 
 class Command_Processor:
-    def __init__(self, argv, pattern_list=None, logfile=None):
+    def __init__(self, argv, pattern_list=None, logfile=None,
+                 stdout=None):
         self.pattern_list = pattern_list + [pexpect.EOF, pexpect.TIMEOUT]
         self.error = None
         self.exit_status = None
+        self.pipe = None
         self.prog = None
 
+        if stdout is not None:
+            self.pipe = subprocess.Popen(argv, stdout=stdout,
+                                         stdin=subprocess.DEVNULL,
+                                         stderr=subprocess.PIPE)
+
         try:
-            self.prog = pexpect.spawn(argv[0], argv[1:],
-                                      encoding='utf-8', codec_errors='replace',
-                                      logfile=logfile)
+
+            if self.pipe is None:
+                self.prog = pexpect.spawn(argv[0], argv[1:],
+                                          encoding='utf-8',
+                                          codec_errors='replace',
+                                          logfile=logfile)
+            else:
+                self.prog = fdpexpect.fdspawn(self.pipe.stderr,
+                                              encoding='utf-8',
+                                              codec_errors='replace',
+                                              logfile=logfile)
         except pexpect.ExceptionPexpect:
             self.error = 'Failed to start command: %s' % sys.exc_info()[1]
             return
@@ -117,8 +129,12 @@ class Command_Processor:
             if got < len(self.pattern_list) - 2:
                 return (got, self.prog.match)
             if got == len(self.pattern_list) - 2:
-                self.prog.close()
-                self.exit_status = self.prog.exitstatus
+                if self.pipe is None:
+                    self.prog.close()
+                    self.exit_status = self.prog.exitstatus
+                else:
+                    self.pipe.wait()
+                    self.exit_status = self.pipe.returncode
                 break   # EOF: done
             if got == len(self.pattern_list) - 1:
                 pass    # timeout; process events and then try again...
@@ -228,7 +244,7 @@ class Preferences_Editor(QDialog, Ui_Preferences_Dialog):
         self.parent = parent
         self.setupUi(self)
 
-    def exec(self):
+    def exec_(self):
         self.sn_service_combo.currentIndexChanged.connect(
             self.sn_service_changed)
         self.sn_service_combo.setCurrentIndex(0)
@@ -241,7 +257,7 @@ class Preferences_Editor(QDialog, Ui_Preferences_Dialog):
                     break
         self.increment_spinbox.setValue(self.parent.preferences.sn_increment)
         self.station_id_spinbox.setValue(self.parent.preferences.station_id)
-        super().exec()
+        super().exec_()
 
     def accept(self):
         increment = self.increment_spinbox.value()
@@ -294,10 +310,6 @@ class UI(QMainWindow, Ui_MainWindow):
         self.current_params = self.params['AC']
         self.current_params.activate()
 
-        appicon = pkg_resources \
-                  .resource_filename(__name__,
-                                     'resources/icons/ctid-logo.png')
-        window.setWindowIcon(QtGui.QIcon(appicon))
         self.auto_serial_checkbox.setChecked(True)
         self.serial_spinbox.setEnabled(False)
         self.console = terminal.Terminal(self.plainTextEdit)
@@ -308,7 +320,7 @@ class UI(QMainWindow, Ui_MainWindow):
         self.mfg_id = None
         self.model = None
 
-        self.console.write('Welcome to the CTid Programmer!\n\n'
+        self.console.write('Welcome to the CTid® Programmer!\n\n'
                            'You can use this tool to program '
                            'the microcontroller of a CTid board (PCB).\n\n'
                            'Please start by filling out the form on the left. '
@@ -387,14 +399,14 @@ class UI(QMainWindow, Ui_MainWindow):
         self.template_list_dialog.operation = 'load'
         self.template_list_dialog.template_name_frame.hide()
         self.template_list_dialog.buttonBox.setStandardButtons(
-            QtGui.QDialogButtonBox.Open|QtGui.QDialogButtonBox.Cancel)
+            QDialogButtonBox.Open|QDialogButtonBox.Cancel)
         self.template_list_dialog.exec()
 
     def template_save(self):
         self.template_list_dialog.operation = 'save'
         self.template_list_dialog.template_name_frame.show()
         self.template_list_dialog.buttonBox.setStandardButtons(
-            QtGui.QDialogButtonBox.Save|QtGui.QDialogButtonBox.Cancel)
+            QDialogButtonBox.Save|QDialogButtonBox.Cancel)
         self.template_list_dialog.exec()
 
     def get_mfg_id(self):
@@ -513,8 +525,9 @@ class UI(QMainWindow, Ui_MainWindow):
         self.read_btn.setEnabled(enabled)
         self.reprogram_after_cal_btn.setEnabled(enabled)
 
-    def cmd_start(self, argv, pattern_list, logfile=None):
-        self.cmd = Command_Processor(argv, pattern_list, logfile=logfile)
+    def cmd_start(self, argv, pattern_list, logfile=None, stdout=None):
+        self.cmd = Command_Processor(argv, pattern_list, logfile=logfile,
+                                     stdout=stdout)
 
     def cmd_done(self):
         if self.cmd.error is not None:
@@ -591,18 +604,24 @@ class UI(QMainWindow, Ui_MainWindow):
             argv += params.encoder_argv(cleaned_data)
 
         output = tempfile.mkstemp(suffix='.hex', prefix='CTid-')
-        argv += ['-o', output[1]]
 
         template_filename = CODE_TEMPLATE[st] if st in CODE_TEMPLATE \
                             else 'powered.hex'
-        argv.append(os.path.join(PATH_CODE_DIR, template_filename))
 
-        log.debug('create_hexfile: argv=%s', argv)
+        ref = (importlib_resources.files('ctid_programmer')
+               / 'resources' / 'code' / template_filename)
+        with importlib_resources.as_file(ref) as code_template_path:
+            argv.append(str(code_template_path))
+
+            log.debug('create_hexfile: argv=%s', argv)
+
+            self.cmd_start(argv, [], logfile=self.console,
+                           stdout=output[0])
+            # consume command's output by iterating over it:
+            for _ in self.cmd:
+                pass
+            self.cmd_done()
 
-        self.cmd_start(argv, [], logfile=self.console)
-        for _ in self.cmd:
-            pass
-        self.cmd_done()
         if self.cmd.exit_status != 0:
             self.show_cmd_error(self.cmd, 'Command Failed',
                                 'Failed to create hex file.')
@@ -655,23 +674,23 @@ class UI(QMainWindow, Ui_MainWindow):
             return
 
     def read_template_from_flash(self, chip_type):
-        temp = tempfile.mkstemp(suffix='.bin', prefix='CTid-')
+        temp = tempfile.TemporaryFile(suffix='.bin', prefix='CTid-')
         self.cmd_start([PATH_AVRDUDE, '-ctc2030', '-p%s' % chip_type,
-                        '-Uflash:r:%s:r' % temp[1]], [], logfile=self.console)
+                        '-Uflash:r:-:r'], [], logfile=self.console,
+                       stdout=temp)
         for _ in self.cmd:
             pass	# consume output until program is done...
         self.cmd_done()
 
         if self.cmd.exit_status != 0:
-            os.close(temp[0])
-            os.remove(temp[1])
+            temp.close()
             self.show_cmd_error(self.cmd, 'Read Failed',
                                 'Failed to read the microcontroller flash.')
             return None
 
-        flash = open(temp[1], 'rb').read()
-        os.close(temp[0])
-        os.remove(temp[1])
+        temp.seek(0)
+        flash = temp.read()
+        temp.close()
 
         if len(flash) < 0x3e1:
             if not flash:
@@ -712,7 +731,9 @@ class UI(QMainWindow, Ui_MainWindow):
         for key, value in cal_params.items():
             msg += '\t%16s: %s\n' % (key, value)
 
-        self.console.write('\n\nFound calibration parameters:\n%s\n' % msg)
+        self.console.write('\n\nFound calibration data for %s %s #%d:\n%s\n' %
+                           (ctid.mfg_short_name(template['mfg']),
+                            template['model'], template['sn'], msg))
 
         # merge the calibrated parameters with existing template:
         for name, value in cal_params.items():
